import aframexr
import unittest

from aframexr.utils.constants import AVAILABLE_ENVIRONMENTS, ERROR_MESSAGES
from .constants import *

START_END = {'start': '0 0 0', 'end': '1 1 1'}


class TestLineOK(unittest.TestCase):
    """Tests for simple line creation."""

    def test_simple(self):
        """Simple line creation."""
        line = aframexr.Line(**START_END)
        line.to_html()

    def test_from_dict(self):
        """Line creation using form_dict() method."""
        line = aframexr.Line.from_dict(aframexr.Line(**START_END).to_dict())
        line.to_html()

    def test_from_json(self):
        """Line creation using form_json() method."""
        line = aframexr.Line.from_json(aframexr.Line(**START_END).to_json())
        line.to_html()

    def test_color(self):
        """Line with specific color creation."""
        for c in SIMPLE_ELEMENTS_COLORS:
            line = aframexr.Line(**START_END, color=c)
            line.to_html()

    def test_concatenation(self):
        """Concatenation of simple lines."""
        concatenated_chart = aframexr.Line(**START_END) + aframexr.Line(**START_END)
        concatenated_chart.to_html()

    def test_environment(self):
        """Scene creation with personalized environment."""
        for e in AVAILABLE_ENVIRONMENTS:
            aframexr.Line(**START_END).to_html(environment=e)

    def test_movable(self):
        """Movable line creation."""
        html = aframexr.Line(**START_END).movable().to_html()
        self.assertIn('movable', html)


class TestLineERROR(unittest.TestCase):
    """Tests for line creation errors."""

    def test_color_error(self):
        """Line with specific color creation error."""
        not_an_string = ['not a color']
        with self.assertRaises(TypeError) as error:
            # noinspection PyTypeChecker
            aframexr.Line(**START_END, color=not_an_string).to_html()
        self.assertEqual(
            str(error.exception),
            ERROR_MESSAGES['TYPE'].format(
                param_name='specs.color', expected_type=str.__name__, current_type=type(not_an_string).__name__
            )
        )

    def test_environment_error(self):
        """Scene creation with invalid personalized environment."""
        environment = 'bad_environment'
        line = aframexr.Line(**START_END)
        with self.assertRaises(ValueError) as error:
            # noinspection PyTypeChecker
            line.to_html(environment=environment)
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENVIRONMENT'].format(environment=environment))



import json

from ..utils.validators import AframeXRValidator


class Data:
    """
    Data class. See the example below to see the format.

    Examples
    --------
    # To instantiate Data using Data(data_format_as_list)

    >>> import aframexr
    >>> data_format_as_list = [{"a": 1, "b": 2}, {"a": 2, "b": 4}]
    >>> data = aframexr.Data(data_format_as_list)

    # To instantiate Data using Data.from_json(data_format_as_string)

    >>> import aframexr
    >>> data_format_as_string = '[{"a": 1, "b": 2}, {"a": 2, "b": 4}]'
    >>> data = aframexr.Data.from_json(data_format_as_string)
    """

    def __init__(self, values: list[dict]):
        AframeXRValidator.validate_type('values', values, list)
        self.values = values

    # Import data
    @staticmethod
    def from_json(data: str):
        """Create a Data object from JSON string."""
        AframeXRValidator.validate_type('data', data, str)
        data = json.loads(data)
        return Data(data)

    # Export data
    def to_json(self) -> str:
        """Return a JSON string representation of the data."""
        return json.dumps(self.values)


class UrlData:
    """
    UrlData class.

    Examples
    --------
    >>> import aframexr
    >>> url = '...'  # The URL of the file storing the data
    >>> data = aframexr.UrlData(url)
    """

    def __init__(self, url: str):
        self.url = url


from ..utils.validators import AframeXRValidator

# Selections
def selection_point(name: str, fields: list) -> 'Parameter':
    """Add selection to the chart."""
    AframeXRValidator.validate_type('name', name, str)
    AframeXRValidator.validate_type('fields', fields, list)

    select_config = {'type': 'point', 'fields': fields}
    return Parameter(name=name, select=select_config)


class Parameter:
    def __init__(self, name: str, select: dict):
        AframeXRValidator.validate_type('name', name, str)
        AframeXRValidator.validate_type('select', select, dict)

        self._name = name
        self._select_config = select

    def to_specs(self):
        return {
            'name': self._name,
            'select': self._select_config
        }

    def to_dict(self) -> dict:
        return {'param': self._name}



import aframexr
import unittest

from aframexr.utils.constants import AVAILABLE_ENVIRONMENTS

from tests.constants import *

URL = 'https://raw.githubusercontent.com/davidlab20/TFG/main/docs/static/imgs/logo.png'


class TestMarkImageOK(unittest.TestCase):
    """Mark image OK tests."""
    def test_simple(self):
        """Simple image creation."""
        aframexr.Image(URL).to_html()

    def test_simple_with_pandas_not_installed(self):
        """Simple image creation without having pandas installed."""
        import sys
        import importlib
        from unittest import mock

        sys.modules.pop('aframexr', None)
        sys.modules.pop('aframexr.api.components', None)

        with mock.patch.dict(sys.modules, {'pandas': None}):
            import aframexr
            importlib.reload(aframexr)

            aframexr.Image(URL).to_html()

            import aframexr.api.components as components
            self.assertIsNone(components.pd)
            self.assertIs(components.DataFrame, object)

    def test_from_json(self):
        """Image using from_json() method creation."""
        json_string = aframexr.Image(URL).to_json()
        image_chart = aframexr.Chart.from_json(json_string)
        image_chart.to_html()

        self.assertTrue(image_chart.to_json() == json_string)

    def test_position(self):
        """Image changing position creation."""
        for p in POSITIONS:
            aframexr.Image(URL, position=p).to_html()

    def test_position_format(self):
        """Image changing position format creation."""
        for p in POSITION_FORMATS:
            aframexr.Image(URL, position=p).to_html()

    def test_rotation(self):
        """Image changing rotation creation."""
        for r in ROTATIONS:
            aframexr.Image(URL, rotation=r).to_html()

    def test_rotation_format(self):
        """Image changing rotation format creation."""
        for r in ROTATION_FORMATS:
            aframexr.Image(URL, rotation=r).to_html()

    def test_width(self):
        """Image changing width creation."""
        for w in ALL_MARK_DEPTHS_HEIGHTS_WIDTHS:
            aframexr.Image(URL, width=w).to_html()

    def test_height(self):
        """Image changing height creation."""
        for h in ALL_MARK_DEPTHS_HEIGHTS_WIDTHS:
            aframexr.Image(URL, height=h).to_html()

    def test_position_rotation_width_height(self):
        """Image changing position, rotation, width and height creation."""
        for p, r, w, h in zip(POSITIONS, ROTATIONS, ALL_MARK_DEPTHS_HEIGHTS_WIDTHS, ALL_MARK_DEPTHS_HEIGHTS_WIDTHS):
            aframexr.Image(URL, position=p, rotation=r, width=w, height=h).to_html()

    def test_concatenation(self):
        """Image concatenation creation."""
        concatenated_chart = aframexr.Image(URL, position=CONCATENATION_POSITIONS[0])
        for pos in CONCATENATION_POSITIONS[1:]:
            concatenated_chart += aframexr.Image(URL, position=pos)

        concatenated_chart.to_html()

    def test_environment(self):
        """Scene creation with personalized environment."""
        for e in AVAILABLE_ENVIRONMENTS:
            aframexr.Image(URL).to_html(environment=e)

    def test_save(self):
        """Image saving."""
        import tempfile
        from pathlib import Path

        with tempfile.TemporaryDirectory() as tmpdir:
            temp_html_file_path = Path(tmpdir) / "test.html"
            temp_json_file_path = Path(tmpdir) / "test.json"

            image = aframexr.Image(URL)
            image.save(str(temp_html_file_path))
            image.save(str(temp_json_file_path))

            assert temp_html_file_path.exists()
            assert temp_json_file_path.exists()

    def test_movable(self):
        """Movable image creation."""
        html = aframexr.Image(URL).movable().to_html()
        self.assertIn('movable', html)



from .constants import ENTITY_IS_MOVABLE

CREATOR_MAP: dict[str, type['ElementCreator']] = {}  # Creator map of elements, classes are added at the end of the file


class ElementCreator:
    _ELEMENT_HTML: str = ''  # Must be defined by child classes
    """
    When defining _ELEMENT_HTML in a subclass:
        * Do NOT add a space before {attributes}.
        * The method get_element_html() automatically adds a leading space if there are attributes.
    """

    def __init__(self, element_specs: dict, filtered_by_params: bool = False):
        self._filtered_by_params = filtered_by_params

        if type(self) is ElementCreator:
            raise TypeError(f'{self.__class__.__name__} is abstract')

        self._attributes = {
            key: value for key, value in element_specs.items()
            if value is not None and key not in {'element', 'environment', 'movable'}
        }

    @classmethod
    def create_object(cls, element_type: str, element_specs: dict) -> 'ElementCreator':
        """
        Returns an ElementCreator instance depending on element specifications.
        Notes
        -----
        Supposing that field "element" exists in element specifications, this method is called from ChartsHTMLCreator.
        """
        if element_type not in CREATOR_MAP:  # pragma: no cover (all classes should be added at the end of this file)
            raise RuntimeError(f'Class for {element_type} was not added to CREATOR_MAP')

        return CREATOR_MAP[element_type](element_specs)

    def get_element_html(self, is_movable: bool = ENTITY_IS_MOVABLE) -> str:
        if self._ELEMENT_HTML == '':  # pragma: no cover (all classes should have inner _ELEMENT_HTML constant defined)
            raise RuntimeError('Attribute _ELEMENT_HTML was not initialized')

        attributes = ''.join(
            f' {key.replace("_", "-")}="{value}"'  # Add space at the beginning (using HTML format)
            if type(self) is not LineCreator else f'{key}: {value}; '  # Lines are created different
            for key, value in self._attributes.items()
        )

        has_info = 'info' in self._attributes and not self._filtered_by_params
        if is_movable:
            attributes += ' movable raycastable'
        elif has_info:
            attributes += ' raycastable'

        return self._ELEMENT_HTML.format(attributes=attributes)


class BoxCreator(ElementCreator):
    _ELEMENT_HTML = '<a-box{attributes} side="double"></a-box>'


class ConeCreator(ElementCreator):
    _ELEMENT_HTML = '<a-cone{attributes}></a-cone>'


class CylinderCreator(ElementCreator):
    _ELEMENT_HTML = '<a-cylinder side="double"{attributes}></a-cylinder>'


class DodecahedronCreator(ElementCreator):
    _ELEMENT_HTML = '<a-dodecahedron{attributes}></a-dodecahedron>'


class GLTFCreator(ElementCreator):
    _ELEMENT_HTML = '<a-gltf-model{attributes}></a-gltf-model>'


class IcosahedronCreator(ElementCreator):
    _ELEMENT_HTML = '<a-icosahedron{attributes}></a-icosahedron>'


class ImageCreator(ElementCreator):
    _ELEMENT_HTML = '<a-image{attributes}></a-image>'

class OctahedronCreator(ElementCreator):
    _ELEMENT_HTML = '<a-octahedron{attributes}></a-octahedron>'

class LineCreator(ElementCreator):
    _ELEMENT_HTML = '<a-entity line="{attributes}"></a-entity>'


class PlaneCreator(ElementCreator):
    _ELEMENT_HTML = '<a-plane side="double"{attributes}></a-plane>'


class SphereCreator(ElementCreator):
    _ELEMENT_HTML = '<a-sphere{attributes}></a-sphere>'


class TetrahedronCreator(ElementCreator):
    _ELEMENT_HTML = '<a-tetrahedron{attributes}></a-tetrahedron>'


class TextCreator(ElementCreator):
    _ELEMENT_HTML = '<a-text{attributes}></a-text>'


class TorusCreator(ElementCreator):
    _ELEMENT_HTML = '<a-torus{attributes}></a-torus>'


# Add creator classes to CREATOR_MAP dynamically
CREATOR_MAP.update({
    'box': BoxCreator,
    'cone': ConeCreator,
    'cylinder': CylinderCreator,
    'dodecahedron': DodecahedronCreator,
    'gltf': GLTFCreator,
    'icosahedron': IcosahedronCreator,
    'image': ImageCreator,
    'octahedron': OctahedronCreator,
    'line': LineCreator,
    'plane': PlaneCreator,
    'sphere': SphereCreator,
    'tetrahedron': TetrahedronCreator,
    'text': TextCreator,
    'torus': TorusCreator,
})



import aframexr
import math
import unittest

from bs4 import BeautifulSoup

from aframexr.api.filters import FilterTransform
from aframexr.utils.constants import AVAILABLE_ENVIRONMENTS, DEFAULT_CHART_HEIGHT, ERROR_MESSAGES
from tests.constants import *  # Constants used for testing


def _bars_bases_are_on_x_axis(bars_chart_html: str) -> bool:
    """Verify that the bars are well-placed in the x-axis (the base of the bar is in the x-axis)."""
    soup = BeautifulSoup(bars_chart_html, 'lxml')
    x_axis_y_pos = float(soup.select('a-entity[line]')[0]['line'].split(';')[0].split()[2])  # Y position of x-axis line

    bars = soup.find_all('a-box')
    for b in bars:
        bar_height = float(b['height'])  # Total height of the bar
        y_axis_midpoint = float(b['position'].split()[1])  # Y-axis coordinate

        y_id = float(b['info'].split(';')[1].split(':')[1])
        if y_id >= 0:  # Bar represents positive value (above x-axis)
            if not math.isclose(x_axis_y_pos, y_axis_midpoint - 0.5 * bar_height):
                print(f'\nDEBUG: Positive bar\'s base is not on x-axis line.'
                      f'\n\t- X-axis line Y-coordinate: {x_axis_y_pos}'
                      f'\n\t- Y-axis bar coordinate: {y_axis_midpoint}'
                      f'\n\t- Bar\'s height: {bar_height}')
                return False  # Y-pos minus half its height must be the same as the x-axis y-coordinate
        else:  # Bar represents negative value (below x-axis)
            if not math.isclose(x_axis_y_pos, y_axis_midpoint + 0.5 * bar_height):
                print(f'\nDEBUG: Negative bar\'s base is not on x-axis line.'
                      f'\n\t- X-axis line Y-coordinate: {x_axis_y_pos}'
                      f'\n\t- Y-axis bar coordinate: {y_axis_midpoint}'
                      f'\n\t- Bar\'s height: {bar_height}')
                return False  # Y-pos plus half its height must be the same as the x-axis y-coordinate
    return True


def _bars_height_does_not_exceed_max_height(bars_chart_html: str, bars_chart: aframexr.Chart) -> bool:
    """Verify that every bar height does not exceed the maximum height."""
    max_height = float(bars_chart.to_dict().get('height', DEFAULT_CHART_HEIGHT))

    soup = BeautifulSoup(bars_chart_html, 'lxml')
    bars = soup.find_all('a-box')
    for b in bars:
        bar_height = float(b['height'])  # Total height of the bar
        if bar_height > max_height:
            print(f'\nDEBUG: Bar\'s height exceed chart\'s height.'
                  f'\n\t- Bar\'s height: {bar_height}'
                  f'\n\t- Chart height: {max_height}')
            return False
    return True


class TestMarkBarOK(unittest.TestCase):
    """Bars chart OK tests."""

    def test_simple(self):
        """Simple bars chart creation."""
        bars_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales')
        bars_chart_html = bars_chart.to_html()
        self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
        self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_simple_with_pandas_not_installed(self):
        """Simple bars chart creation without having pandas installed."""
        import sys
        import importlib
        from unittest import mock

        sys.modules.pop('aframexr', None)
        sys.modules.pop('aframexr.api.components', None)

        with mock.patch.dict(sys.modules, {'pandas': None}):
            import aframexr
            importlib.reload(aframexr)

            bars_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales')
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

            import aframexr.api.components as components
            self.assertIsNone(components.pd)
            self.assertIs(components.DataFrame, object)

    def test_from_json(self):
        """Bars chart using from_json() method creation."""
        json_string = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales').to_json()
        bars_chart = aframexr.Chart.from_json(json_string)
        bars_chart_html = bars_chart.to_html()

        self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
        self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))  # type: ignore
        self.assertEqual(bars_chart.to_json(), json_string)

    def test_data_format(self):
        """Bars chart changing data format creation."""
        for d in DATA_FORMATS:
            bars_chart = aframexr.Chart(d).mark_bar().encode(x='model', y='sales')
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_position(self):
        """Bars chart changing position creation."""
        for p in POSITIONS:
            bars_chart = aframexr.Chart(DATA, position=p).mark_bar().encode(x='model', y='sales')
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_position_format(self):
        """Bars chart changing position format creation."""
        for p in POSITION_FORMATS:
            bars_chart = aframexr.Chart(DATA, position=p).mark_bar().encode(x='model', y='sales')
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_rotation(self):
        """Bars chart changing rotation creation."""
        for r in ROTATIONS:
            bars_chart = aframexr.Chart(DATA, rotation=r).mark_bar().encode(x='model', y='sales')
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_rotation_format(self):
        """Bars chart changing rotation format creation."""
        for r in ROTATION_FORMATS:
            bars_chart = aframexr.Chart(DATA, rotation=r).mark_bar().encode(x='model', y='sales')
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_depth(self):
        """Bars chart changing depth creation."""
        for d in ALL_MARK_DEPTHS_HEIGHTS_WIDTHS:
            bars_chart = aframexr.Chart(DATA, depth=d).mark_bar().encode(x='model', y='sales')
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_height(self):
        """Bars chart changing height creation."""
        for h in MARK_BAR_POINT_HEIGHTS_WIDTHS:
            bars_chart = aframexr.Chart(DATA, height=h).mark_bar().encode(x='model', y='sales')
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_width(self):
        """Bars chart changing width creation."""
        for w in MARK_BAR_POINT_HEIGHTS_WIDTHS:
            bars_chart = aframexr.Chart(DATA, width=w).mark_bar().encode(x='model', y='sales')
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_title(self):
        """Bar chart with title creation."""
        bars_chart = aframexr.Chart(DATA, title='Title').mark_bar().encode(x='model', y='sales')
        bars_chart_html = bars_chart.to_html()
        self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
        self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_color(self):
        """Bars chart changing color creation."""
        for c in SIMPLE_ELEMENTS_COLORS:
            bars_chart = aframexr.Chart(DATA).mark_bar(color=c).encode(x='model', y='sales')
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_size(self):
        """Bars chart changing size creation."""
        for s in MARK_BAR_POINT_SIZES:
            bars_chart = aframexr.Chart(DATA).mark_bar(size=s).encode(x='model', y='sales')
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_encoding(self):
        """Bars chart changing encoding creation."""
        for e in MARK_BAR_ENCODINGS:
            bars_chart = aframexr.Chart(DATA).mark_bar().encode(**e)
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_encoding_with_no_Y_axis_displayed(self):
        """Bar chart creation with no Y-axis displayed."""
        bars_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y=aframexr.Y('sales', axis=None))
        bars_chart_html = bars_chart.to_html()
        self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
        self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_filter(self):
        """Bars chart changing filter creation."""
        for eq in FILTER_EQUATIONS:
            for f in [eq, FilterTransform.from_equation(eq)]:  # Filter using equation and using FilterTransform object
                bars_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales').transform_filter(f)
                bars_chart_html = bars_chart.to_html()
                self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
                self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_dynamic_filter(self):
        """Bars chart with dynamic filter creation."""
        bars_chart_1 = aframexr.Chart(DATA).mark_bar().encode(x='motor', y='sales').add_params(DYNAMIC_FILTER)
        bars_chart_2 = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales').transform_filter(DYNAMIC_FILTER)
        final_chart = bars_chart_1 + bars_chart_2
        final_chart_html = final_chart.to_html()
        self.assertTrue(_bars_bases_are_on_x_axis(final_chart_html))
        self.assertTrue(_bars_height_does_not_exceed_max_height(final_chart_html, final_chart))

    def test_several_filters(self):
        """Bars chart with several filters' creation."""
        bars_chart = (aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales')
                      .transform_filter(SEVERAL_FILTER_EQUATIONS[0]))
        for eq in SEVERAL_FILTER_EQUATIONS[1:]:
            bars_chart.transform_filter(eq)

        bars_chart_html = bars_chart.to_html()
        self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
        self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_aggregate(self):
        """Bar chart changing aggregates creation."""
        for a in AGGREGATES:
            bars_chart = (aframexr.Chart(DATA).mark_bar().encode(x='model', y='new_field')
                          .transform_aggregate(new_field=f'{a}(sales)'))
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

            bars_chart_2 = aframexr.Chart(DATA).mark_bar().encode(x='model', y=f'{a}(sales)')
            bars_chart_2_html = bars_chart_2.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_2_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_2_html, bars_chart_2))

    def test_aggregate_position_rotation_size_height_width_filter(self):
        """Bars chart changing position, rotation size, height, width and filter creation."""
        for a, p, r, s, h, w, f in zip(AGGREGATES, POSITIONS, ROTATIONS, MARK_BAR_POINT_SIZES,
                                       MARK_BAR_POINT_HEIGHTS_WIDTHS, MARK_BAR_POINT_HEIGHTS_WIDTHS,
                                       FILTER_EQUATIONS):
            bars_chart = (aframexr.Chart(DATA, position=p, rotation=r, height=h, width=w).mark_bar(size=s)
                          .encode(x='model', y='agg').transform_filter(f).transform_aggregate(agg=f'{a}(sales)'))
            bars_chart_html = bars_chart.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_concatenation(self):
        """Bars chart concatenation creation."""
        concatenated_chart = (aframexr.Chart(DATA, position=CONCATENATION_POSITIONS[0]).mark_bar()
                              .encode(x='model', y='sales'))
        for pos in CONCATENATION_POSITIONS[1:]:
            concatenated_chart += aframexr.Chart(DATA, position=pos).mark_bar().encode(x='model', y='sales')

        concatenated_chart_html = concatenated_chart.to_html()
        self.assertTrue(_bars_bases_are_on_x_axis(concatenated_chart_html))
        self.assertTrue(_bars_height_does_not_exceed_max_height(concatenated_chart_html, concatenated_chart))

    def test_environment(self):
        """Scene creation with personalized environment."""
        for e in AVAILABLE_ENVIRONMENTS:
            bars_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales')
            bars_chart_html = bars_chart.to_html(environment=e)
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))

    def test_save(self):
        """Bars chart saving."""
        import tempfile
        from pathlib import Path

        with tempfile.TemporaryDirectory() as tmpdir:
            temp_html_file_path = Path(tmpdir) / "test.html"
            temp_json_file_path = Path(tmpdir) / "test.json"

            bars_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales')
            bars_chart.save(str(temp_html_file_path))
            bars_chart.save(str(temp_json_file_path))

            self.assertTrue(temp_html_file_path.exists())
            self.assertTrue(temp_json_file_path.exists())

    def test_properties(self):
        """Bars chart properties definition."""
        bars_chart = aframexr.Chart().mark_bar().encode(x='model', y='sales')
        for p, r in zip(POSITIONS, ROTATIONS):
            bars_chart_2 = bars_chart.properties(data=DATA, position=p, rotation=r)
            bars_chart_2_html = bars_chart_2.to_html()
            self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_2_html))
            self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_2_html, bars_chart_2))

    def test_movable(self):
        """Movable bar chart creation."""
        bars_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales').movable()
        bars_chart_html = bars_chart.to_html()
        self.assertIn('movable', bars_chart_html)
        self.assertTrue(_bars_bases_are_on_x_axis(bars_chart_html))
        self.assertTrue(_bars_height_does_not_exceed_max_height(bars_chart_html, bars_chart))


class TestMarkBarError(unittest.TestCase):
    """Bars chart error tests."""

    def test_load_data_url_error(self):
        """Bars chart load data url error."""
        with self.assertRaises(IOError) as error:
            aframexr.Chart(NON_EXISTING_URL_DATA).mark_bar().encode(x='model', y='sales').to_html()

        self.assertEqual(str(error.exception), f"Could not load data from URL: {NON_EXISTING_URL_DATA.url}.")

    def test_local_file_does_not_exist(self):
        """Bars chart error when the local file does not exist."""
        with self.assertRaises(IOError) as error:
            aframexr.Chart(NON_EXISTING_LOCAL_PATH).mark_bar().encode(x='model', y='sales').to_html()

        self.assertRegex(str(error.exception), r'Local file .* was not found')

    def test_bad_file_format(self):
        """Bars chart error when the file format is incorrect."""
        with open(BAD_FILE_FORMAT.url, 'w'):
            pass  # Create a bad format temporal file

        try:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(BAD_FILE_FORMAT).mark_bar().encode(x='model', y='sales').to_html()

            self.assertIn(f'Unsupported file type: ', str(error.exception))

        finally:
            from pathlib import Path
            Path(BAD_FILE_FORMAT.url).unlink()  # Remove the temporal file

    def test_file_is_empty(self):
        """Bars chart error when the file is empty."""
        with open(EMPTY_FILE.url, 'w'):
            pass  # Create an empty temporal file

        try:
            with self.assertRaises(IOError) as error:
                aframexr.Chart(EMPTY_FILE).mark_bar().encode(x='model', y='sales').to_html()

            self.assertIn('Error when processing data. Error: ', str(error.exception))

        finally:
            from pathlib import Path
            Path(EMPTY_FILE.url).unlink()  # Remove the temporal file

    def test_color_encoding_not_nominal_error(self):
        """Bar chart error when color encoding is not nominal."""
        with self.assertRaises(ValueError) as error:
            aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales', color='doors').to_html()  # Doors, quantitative
        self.assertEqual(
            str(error.exception),
            ERROR_MESSAGES['COLOR_ENCODING_NOT_NOMINAL'].format(color_encoding='quantitative')
        )

    def test_position_error(self):
        """Bars chart position error."""
        for p in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_bar().encode(x='model', y='sales').to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['NOT_3_AXES_POSITION_OR_ROTATION'].format(
                pos_or_rot='position', pos_or_rot_value=p
            ))

        for p in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, position=p).mark_bar().encode(x='model', y='sales').to_html()
            self.assertEqual(str(error.exception), 'The position values must be numeric.')

    def test_rotation_error(self):
        """Bars chart rotation error."""
        for r in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_bar().encode(x='model', y='sales').to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['NOT_3_AXES_POSITION_OR_ROTATION'].format(
                pos_or_rot='rotation', pos_or_rot_value=r
            ))

        for r in NOT_NUMERIC_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, rotation=r).mark_bar().encode(x='model', y='sales').to_html()
            self.assertEqual(str(error.exception), 'The rotation values must be numeric.')

    def test_size_error(self):
        """Bars chart size error."""
        for s in NOT_GREATER_THAN_0_NUMBERS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_bar(size=s).encode(x='model', y='sales')
            self.assertEqual(str(error.exception), ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='size'))

    def test_depth_error(self):
        """Bars chart error when depth is incorrect."""
        for d in NOT_GREATER_THAN_0_NUMBERS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, depth=d).mark_bar().encode(x='model', y='sales').to_html()
            self.assertEqual(
                str(error.exception),
                ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='depth')
            )

    def test_height_error(self):
        """Bars chart height error."""
        for h in NOT_GREATER_THAN_0_NUMBERS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, height=h).mark_bar().encode(x='model', y='sales').to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='height'))

    def test_width_error(self):
        """Bars chart width error."""
        for w in NOT_GREATER_THAN_0_NUMBERS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA, width=w).mark_bar().encode(x='model', y='sales').to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='width'))

    def test_encoding_error(self):
        """Bars chart encoding error."""
        for e in NON_EXISTING_MARK_BAR_POINT_ENCODINGS:
            with self.assertRaises(KeyError) as error:
                bars_chart = aframexr.Chart(DATA).mark_bar().encode(**e)
                bars_chart.to_html()
            self.assertIn('Data has no field ', str(error.exception))

        for e in NOT_VALID_MARK_BAR_POINT_ENCODINGS:
            with self.assertRaises(ValueError) as error:
                aframexr.Chart(DATA).mark_bar().encode(**e).to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['LESS_THAN_2_XYZ_ENCODING'])

    def test_encoding_error_invalid_encoding_type(self):
        """Bars chart encoding error with invalid encoding type."""
        with self.assertRaises(ValueError) as error:
            bad_encoding_type = 'BAD_ENCODING'
            aframexr.Chart(DATA).mark_bar().encode(x=f'model:{bad_encoding_type}', y='sales').to_html()
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENCODING_TYPE'].format(encoding_type=bad_encoding_type))

    def test_encoding_error_not_encoded(self):
        """Bars chart encoding error. Encoding not in specifications."""
        with self.assertRaises(ValueError) as error:
            bars_chart = aframexr.Chart(DATA).mark_bar()
            bars_chart.to_html()
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENCODING_NOT_IN_SPECS'])

    def test_filter_warning(self):
        """Bars chart filter warning."""
        for f in WARNING_FILTER_EQUATIONS:
            with self.assertWarns(UserWarning) as warning:
                filt_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales').transform_filter(f)
                filt_chart.to_html()
            self.assertEqual(
                str(warning.warning),
                f'Data does not contain values for the filter: {FilterTransform.from_equation(f).to_dict()}'
            )

    def test_filter_error(self):
        """Bars chart filter error."""
        for f in SYNTAX_ERROR_FILTER_EQUATIONS:
            with self.assertRaises(SyntaxError) as error:
                filt_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales').transform_filter(f)
                filt_chart.to_html()
            self.assertEqual(str(error.exception), 'Incorrect syntax, must be datum.{field} {operator} {value}')

    def test_aggregate_error(self):
        """Bars chart aggregate error."""
        for a in NOT_VALID_AGGREGATES:
            with self.assertRaises(ValueError) as error:
                agg_chart = (aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales')
                             .transform_aggregate(new_field=f'{a}(sales)'))
                agg_chart.to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['AGGREGATE_OPERATION'].format(operation=a))

    def test_aggregate_error_not_defined_encoding_channels(self):
        """Bars chart aggregate error raised when encoding channels are not defined in aggregation."""
        with self.assertRaises(ValueError) as error:
            agg_chart = (aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales')
                         .transform_aggregate(mean_sales='mean(sales)', groupby=['model']))
            agg_chart.to_html()

        self.assertRegex(str(error.exception), r'Encoding channel\(s\) .* must be defined in aggregate .*'
                                               r', otherwise that fields will disappear.')

    def test_environment_error(self):
        """Scene creation with invalid personalized environment."""
        environment = 'bad_environment'
        bars_chart = aframexr.Chart(DATA).mark_bar().encode(x='model', y='sales')
        with self.assertRaises(ValueError) as error:
            # noinspection PyTypeChecker
            bars_chart.to_html(environment=environment)
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENVIRONMENT'].format(environment=environment))



import aframexr
import unittest

from aframexr.utils.constants import AVAILABLE_ENVIRONMENTS, ERROR_MESSAGES
from .constants import *


class TestSphereOK(unittest.TestCase):
    """Tests for simple sphere creation."""

    def test_simple(self):
        """Simple sphere creation."""
        sphere = aframexr.Sphere()
        sphere.to_html()

    def test_from_dict(self):
        """Sphere creation using form_dict() method."""
        sphere = aframexr.Sphere.from_dict(aframexr.Sphere().to_dict())
        sphere.to_html()

    def test_from_json(self):
        """Sphere creation using form_json() method."""
        sphere = aframexr.Sphere.from_json(aframexr.Sphere().to_json())
        sphere.to_html()

    def test_color(self):
        """Sphere with specific color creation."""
        for c in SIMPLE_ELEMENTS_COLORS:
            sphere = aframexr.Sphere(color=c)
            sphere.to_html()

    def test_position(self):
        """Sphere with specific position creation."""
        for p in POSITIONS + POSITION_FORMATS:
            sphere = aframexr.Sphere(position=p)
            sphere.to_html()

    def test_radius(self):
        """Sphere with specific radius creation."""
        for r in RADIUS:
            sphere = aframexr.Sphere(radius=r)
            sphere.to_html()

    def test_all_parameters(self):
        """Sphere creation with all parameters specified."""
        for c, p, r in zip(SIMPLE_ELEMENTS_COLORS, POSITIONS, RADIUS):
            sphere = aframexr.Sphere(color=c, position=p, radius=r)
            sphere.to_html()

    def test_concatenation(self):
        """Concatenation of simple spheres."""
        concatenated_chart = aframexr.Sphere(position=CONCATENATION_POSITIONS[0])
        for p in CONCATENATION_POSITIONS[1:]:
            concatenated_chart += aframexr.Sphere(position=p)
        concatenated_chart.to_html()

    def test_environment(self):
        """Scene creation with personalized environment."""
        for e in AVAILABLE_ENVIRONMENTS:
            aframexr.Sphere().to_html(environment=e)

    def test_movable(self):
        """Movable sphere creation."""
        html = aframexr.Sphere().movable().to_html()
        self.assertIn('movable', html)


class TestSphereERROR(unittest.TestCase):
    """Tests for sphere creation error."""

    def test_color_error(self):
        """Sphere with specific color creation error."""
        not_an_string = ['not a color']
        with self.assertRaises(TypeError) as error:
            # noinspection PyTypeChecker
            sphere = aframexr.Sphere(color=not_an_string)
            sphere.to_html()
        self.assertEqual(
            str(error.exception),
            ERROR_MESSAGES['TYPE'].format(
                param_name='specs.color', expected_type=str.__name__, current_type=type(not_an_string).__name__
            )
        )

    def test_radius_error(self):
        """Sphere with specific radius creation error."""
        for r in NOT_GREATER_THAN_0_RADIUS:
            with self.assertRaises(ValueError) as error:
                sphere = aframexr.Sphere(radius=r)
                sphere.to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='specs.radius'))

    def test_environment_error(self):
        """Scene creation with invalid personalized environment."""
        environment = 'bad_environment'
        sphere = aframexr.Sphere()
        with self.assertRaises(ValueError) as error:
            # noinspection PyTypeChecker
            sphere.to_html(environment=environment)
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENVIRONMENT'].format(environment=environment))



"""Constant / default values utils file"""

# ----- CONSTANTS -----
AVAILABLE_AGGREGATES = {'count', 'max', 'median', 'mean', 'min', 'std', 'sum', 'var'}
AVAILABLE_COLORS = ['red', 'green', 'blue', 'yellow', 'magenta', 'cyan']  # Using list to maintain order
AVAILABLE_ENCODING_TYPES = {'Q': 'quantitative', 'N': 'nominal'}
AVAILABLE_ENVIRONMENTS = {'default', 'contact', 'egypt', 'checkerboard', 'forest', 'goaland', 'yavapai', 'goldmine',
                          'arches', 'threetowers', 'poison', 'tron', 'japan', 'dream', 'volcano', 'starry', 'osiris'}
AVAILABLE_MARKS = {'arc', 'bar', 'line', 'point'}

ENTITY_IS_MOVABLE = False

EPSILON = 1e-5  # To avoid floating problems

START_LABEL_OFFSET = 0.25  # Offset for the start label of the axis
X_LABELS_Z_DELTA = 0.5  # Variation in the y-axis between the labels and the axis (add to x-axis pos for label pos)
LABELS_X_DELTA = -0.5  # Variation in the x-axis between the labels and the axis (add to y and z axis pos for label pos)
LABELS_Y_DELTA = 0.01  # Variation in the y-axis between the labels and the axis (add to x and z axis pos for label pos)

LABELS_SCALE = '1.5 1.5 1.5'

LEGEND_WIDTH = 3  # Width of the legend
LEGEND_HEIGHT_PER_ELEMENT = 0.6  # Legend's height per element in the legend

PLANE_TITLE_HEIGHT = 1  # Height of the title's plane
PLANE_TITLE_SEPARATION = 0.5  # Separation between the chart and the title
TITLE_TEXT_COLOR = 'black'
TITLE_TEXT_SCALE = '2 2 2'

# ----- DEFAULTS -----
# General
DEFAULT_CHART_POS = '0 0 0'  # Default position of the chart
DEFAULT_CHART_ROTATION = '0 0 0'  # Default chart rotation
DEFAULT_CHART_DEPTH = 2  # Default depth of the chart
DEFAULT_CHART_HEIGHT = 4  # Default height of the chart
DEFAULT_CHART_WIDTH = 4  # Default width of the chart

DEFAULT_ELEMENTS_COLOR_IN_CHART = 'blue'  # Default elements color in chart

DEFAULT_NUM_OF_TICKS_IF_QUANTITATIVE_AXIS = 5  # Number of ticks in the axis if it is quantitative

# Bar chart
DEFAULT_BAR_PADDING = 0.2  # Default padding in bar chart

# Line chart
DEFAULT_VERTICES_POINT_DISPLAY = False  # Default points' display in line's vertices
DEFAULT_VERTICES_POINT_VOLUME = 0.002  # Default points' volume in vertices (if displayed)
DEFAULT_VERTICES_SPACING = 0.5  # Default spacing between vertices of the line

# Pie chart
DEFAULT_PIE_RADIUS = 1  # Default radius of the pie chart
DEFAULT_PIE_ROTATION = '-90 0 0'  # Default pie chart rotation

# Point chart
DEFAULT_POINT_VOLUME = 0.15  # Default point's volume

# ========== ERROR MESSAGES ==========
ERROR_MESSAGES = {
    'ALIGN': "Invalid align property: {align}. Must be one of ['center', 'left', 'right']",
    'AGGREGATE_OPERATION': 'Invalid aggregate operation: {operation}',
    'AGGREGATE_OPERATION_NOT_IN_AGGREGATE': 'Aggregate must contain key "op"',
    'COLOR_ENCODING_NOT_NOMINAL': 'Color encoding type must be nominal, got "{color_encoding}"',
    'DATA_WITH_VALUES_AND_URL_IN_SPECS': 'Data cannot contain both "values" and "url"; they are mutually exclusive',
    'DATA_WITH_NOT_VALUES_NEITHER_URL_IN_SPECS': 'Data must contain key "values" or "url"',
    'DATA_NOT_IN_SPECS': 'Invalid chart specifications. Must contain key "data"',
    'ELEMENT_TYPE': 'Invalid element type: {element}',
    'ENCODING_NOT_IN_SPECS': 'Invalid chart specifications. Must contain key "encoding"',
    'ENCODING_TYPE': 'Invalid encoding type: {encoding_type}',
    'ENVIRONMENT': 'Invalid environment: {environment}',
    'LESS_THAN_2_XYZ_ENCODING': 'At least 2 of (x, y, z) must be specified when encoding "mark_bar" or "mark_point"',
    'MARK_AND_ELEMENT_IN_SPECS': 'Specifications cannot contain both "mark" and "element"; they are mutually exclusive',
    'MARK_AND_ELEMENT_NOT_IN_SPECS': 'Invalid chart specifications. Must contain key "mark" or "element"',
    'MARK_TYPE': 'Invalid mark type: {mark_type}',
    'NAME_NOT_IN_PARAM': 'Param specs must contain key "name"',
    'NOT_3_AXES_POSITION_OR_ROTATION': 'The {pos_or_rot}: {pos_or_rot_value} is not correct. Must be "x y z"',
    'NOT_ALL_DATA_VALUES_ARE_DICT': 'Data field "values" must be a list of dictionaries',
    'NOT_ALL_ENCODINGS_ARE_DICT': 'Encoding channels must be dictionaries',
    'PARAM_NOT_SPECIFIED_IN_MARK_ARC': 'Parameter "{param}" must be specified in arc chart',
    'POSITIVE_NUMBER': 'The "{param_name}" must be greater than 0.',
    'SIZE_ENCODING_NOT_QUANTITATIVE': 'Size encoding type must be quantitative, got "{size_encoding}"',
    'TRANSFORM_TYPE': 'Invalid transform type: {transform_type}',
    'TYPE': 'Expected "{param_name}" to be {expected_type}, got {current_type} instead',
}



import aframexr
import unittest
from typing import Literal, get_args

from aframexr.utils.constants import AVAILABLE_ENVIRONMENTS, ERROR_MESSAGES
from .constants import *


ALIGNS = Literal['center', 'left', 'right']
TEXT_CONTENT = 'aframexr'


class TestTextOK(unittest.TestCase):
    """Tests for text creation."""

    def test_simple(self):
        """Simple text creation."""
        text = aframexr.Text(value=TEXT_CONTENT)
        text.to_html()

    def test_from_dict(self):
        """Text creation using form_dict() method."""
        text = aframexr.Text.from_dict(aframexr.Text(value=TEXT_CONTENT).to_dict())
        text.to_html()

    def test_from_json(self):
        """Text creation using form_json() method."""
        text = aframexr.Text.from_json(aframexr.Text(value=TEXT_CONTENT).to_json())
        text.to_html()

    def test_align(self):
        """Text with specific align creation."""
        for a in get_args(ALIGNS):
            text = aframexr.Text(value=TEXT_CONTENT, align=a)
            text.to_html()

    def test_color(self):
        """Text with specific color creation."""
        for c in SIMPLE_ELEMENTS_COLORS:
            text = aframexr.Text(value=TEXT_CONTENT, color=c)
            text.to_html()

    def test_position(self):
        """Text with specific position creation."""
        for p in POSITIONS + POSITION_FORMATS:
            text = aframexr.Text(value=TEXT_CONTENT, position=p)
            text.to_html()

    def test_rotation(self):
        """Text with specific rotation creation."""
        for r in ROTATIONS + ROTATION_FORMATS:
            text = aframexr.Text(value=TEXT_CONTENT, rotation=r)
            text.to_html()

    def test_scale(self):
        """Text with specific scale creation."""
        for s in MARK_GLTF_SCALES:
            text = aframexr.Text(value=TEXT_CONTENT, scale=s)
            text.to_html()

    def test_all_parameters(self):
        """Text creation with all parameters specified."""
        for a, c, p, r, s in zip(get_args(ALIGNS), SIMPLE_ELEMENTS_COLORS, POSITIONS, ROTATIONS, MARK_GLTF_SCALES):
            text = aframexr.Text(value=TEXT_CONTENT, align=a, color=c, position=p, rotation=r, scale=s)
            text.to_html()

    def test_concatenation(self):
        """Concatenation of simple texts."""
        concatenated_chart = aframexr.Text(value=TEXT_CONTENT, position=CONCATENATION_POSITIONS[0])
        for p in CONCATENATION_POSITIONS[1:]:
            concatenated_chart += aframexr.Text(value=TEXT_CONTENT, position=p)
        concatenated_chart.to_html()

    def test_environment(self):
        """Scene creation with personalized environment."""
        for e in AVAILABLE_ENVIRONMENTS:
            aframexr.Text(value=TEXT_CONTENT).to_html(environment=e)

    def test_movable(self):
        """Movable text creation."""
        html = aframexr.Text(value=TEXT_CONTENT).movable().to_html()
        self.assertIn('movable', html)


class TestTextERROR(unittest.TestCase):
    """Tests for text creation error."""

    def test_align_error(self):
        """Text with specific align creation error."""
        incorrect_align = 'not_correct'
        with self.assertRaises(ValueError) as error:
            # noinspection PyTypeChecker
            text = aframexr.Text(value=TEXT_CONTENT, align=incorrect_align)
            text.to_html()
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ALIGN'].format(align=incorrect_align))

    def test_color_error(self):
        """Text with specific color creation error."""
        not_an_string = ['not a color']
        with self.assertRaises(TypeError) as error:
            # noinspection PyTypeChecker
            text = aframexr.Text(value=TEXT_CONTENT, color=not_an_string)
            text.to_html()
        self.assertEqual(
            str(error.exception),
            ERROR_MESSAGES['TYPE'].format(
                param_name='specs.color', expected_type=str.__name__, current_type=type(not_an_string).__name__
            )
        )

    def test_scale_error(self):
        """Text with specific scale creation error."""
        for s in NOT_3AXIS_POSITIONS_ROTATIONS:
            with self.assertRaises(ValueError) as error:
                text = aframexr.Text(value=TEXT_CONTENT, scale=s)
                text.to_html()
            self.assertEqual(
                str(error.exception),
                ERROR_MESSAGES['NOT_3_AXES_POSITION_OR_ROTATION'].format(
                    pos_or_rot='specs.scale', pos_or_rot_value=s
                )
            )

    def test_environment_error(self):
        """Scene creation with invalid personalized environment."""
        environment = 'bad_environment'
        text = aframexr.Text(value=TEXT_CONTENT)
        with self.assertRaises(ValueError) as error:
            # noinspection PyTypeChecker
            text.to_html(environment=environment)
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENVIRONMENT'].format(environment=environment))



from ..api.data import Data

def cars():  # pragma: no cover (as this data is like any other data)
    return Data([
        {"model": "leon", "motor": "electric", "color": "red", "doors": 5, "sales": 10},
        {"model": "ibiza", "motor": "electric", "color": "white", "doors": 3, "sales": 15},
        {"model": "cordoba", "motor": "diesel", "color": "black", "doors": 5, "sales": 3},
        {"model": "toledo", "motor": "diesel", "color": "white", "doors": 5, "sales": 18},
        {"model": "altea", "motor": "diesel", "color": "red", "doors": 5, "sales": 4},
        {"model": "arosa", "motor": "electric", "color": "red", "doors": 3, "sales": 12},
        {"model": "alhambra", "motor": "diesel", "color": "white", "doors": 5, "sales": 5},
        {"model": "600", "motor": "gasoline", "color": "yellow", "doors": 3, "sales": 20},
        {"model": "127", "motor": "gasoline", "color": "white", "doors": 5, "sales": 2},
        {"model": "panda", "motor": "gasoline", "color": "black", "doors": 3, "sales": 13}
    ])

def energy():  # pragma: no cover (as this data is like any other data)
    return Data([
        {"source": "wind", "region": "north", "month": "jan", "production": 700, "cost": 60},
        {"source": "hydro", "region": "west", "month": "jan", "production": 800, "cost": 55},
        {"source": "solar", "region": "south", "month": "jan", "production": 500, "cost": 50},

        {"source": "solar", "region": "north", "month": "feb", "production": 450, "cost": 55},
        {"source": "hydro", "region": "north", "month": "feb", "production": 780, "cost": 57},
        {"source": "wind", "region": "south", "month": "feb", "production": 650, "cost": 58},

        {"source": "wind", "region": "east", "month": "mar", "production": 720, "cost": 62},
        {"source": "hydro", "region": "south", "month": "mar", "production": 820, "cost": 59},
        {"source": "solar", "region": "east", "month": "mar", "production": 480, "cost": 53},

        {"source": "wind", "region": "west", "month": "apr", "production": 750, "cost": 65},
        {"source": "solar", "region": "west", "month": "apr", "production": 520, "cost": 51},
        {"source": "hydro", "region": "east", "month": "apr", "production": 900, "cost": 60}
    ])

def hotels():  # pragma: no cover (as this data is like any other data)
    return Data([
        {"city": "valencia", "type": "hotel", "rooms": 100, "occupancy": 0.82, "price": 130},
        {"city": "madrid", "type": "hotel", "rooms": 120, "occupancy": 0.80, "price": 140},
        {"city": "sevilla", "type": "hotel", "rooms": 80, "occupancy": 0.80, "price": 100},
        {"city": "barcelona", "type": "hotel", "rooms": 130, "occupancy": 0.78, "price": 150},
        {"city": "madrid", "type": "hostel", "rooms": 50, "occupancy": 0.65, "price": 60},
        {"city": "valencia", "type": "hostel", "rooms": 40, "occupancy": 0.60, "price": 55},
        {"city": "barcelona", "type": "hostel", "rooms": 70, "occupancy": 0.72, "price": 75},
        {"city": "madrid", "type": "hotel", "rooms": 100, "occupancy": 0.75, "price": 120},
        {"city": "barcelona", "type": "hotel", "rooms": 150, "occupancy": 0.85, "price": 160},
        {"city": "valencia", "type": "hotel", "rooms": 90, "occupancy": 0.78, "price": 110},
        {"city": "sevilla", "type": "hostel", "rooms": 35, "occupancy": 0.70, "price": 50}
    ])

def orders():  # pragma: no cover (as this data is like any other data)
    return Data([
        {"customer": "c2", "country": "fr", "status": "cancelled", "total": 60},
        {"customer": "c1", "country": "es", "status": "delivered", "total": 140},
        {"customer": "c3", "country": "fr", "status": "pending", "total": 90},
        {"customer": "c4", "country": "de", "status": "processing", "total": 110},
        {"customer": "c2", "country": "fr", "status": "delivered", "total": 200},
        {"customer": "c1", "country": "es", "status": "pending", "total": 80},
        {"customer": "c3", "country": "fr", "status": "delivered", "total": 150},
        {"customer": "c4", "country": "de", "status": "returned", "total": 50},
        {"customer": "c1", "country": "es", "status": "delivered", "total": 120}
    ])

def retail():  # pragma: no cover (as this data is like any other data)
    return Data([
        {"store": "s4", "product": "shirt", "category": "clothing", "units": 15, "revenue": 450},
        {"store": "s1", "product": "phone", "category": "tech", "units": 8, "revenue": 6400},
        {"store": "s3", "product": "desk", "category": "home", "units": 4, "revenue": 1200},
        {"store": "s2", "product": "chair", "category": "home", "units": 10, "revenue": 1500},
        {"store": "s1", "product": "chair", "category": "home", "units": 7, "revenue": 1050},
        {"store": "s4", "product": "shoes", "category": "clothing", "units": 10, "revenue": 600},
        {"store": "s3", "product": "phone", "category": "tech", "units": 6, "revenue": 4800},
        {"store": "s2", "product": "laptop", "category": "tech", "units": 3, "revenue": 3000},
        {"store": "s1", "product": "laptop", "category": "tech", "units": 5, "revenue": 5000}
    ])

def usage():  # pragma: no cover (as this data is like any other data)
    return Data([
        {"user": "u3", "platform": "ios", "country": "es", "minutes": 20},
        {"user": "u2", "platform": "android", "country": "fr", "minutes": 40},
        {"user": "u5", "platform": "ios", "country": "it", "minutes": 50},
        {"user": "u1", "platform": "ios", "country": "es", "minutes": 25},
        {"user": "u4", "platform": "web", "country": "de", "minutes": 200},
        {"user": "u3", "platform": "android", "country": "fr", "minutes": 15},
        {"user": "u2", "platform": "android", "country": "fr", "minutes": 35},
        {"user": "u1", "platform": "ios", "country": "es", "minutes": 30}
    ])




import aframexr
import unittest

from aframexr.utils.validators import ERROR_MESSAGES


class TestAframexrError(unittest.TestCase):
    """General ERROR tests."""
    def test_from_dict_error_validate_type(self):
        """Verify that the error is raised when using from_dict with no dictionary."""
        invalid_specs = 'not_a_dict'
        with self.assertRaises(TypeError) as error:
            # noinspection PyTypeChecker
            aframexr.Chart.from_dict(specs=invalid_specs)
        self.assertEqual(
            str(error.exception),
            ERROR_MESSAGES['TYPE'].format(
                param_name='specs', expected_type=dict.__name__, current_type=type(invalid_specs).__name__
            )
        )

    def test_data_has_values_and_url(self):
        """Verify that the error is raised when having fields "values" and "url" in data."""
        bad_data_specifications = {'data': {'url': '', 'values': ''}, 'mark': 'bar', 'encoding': {}}
        with self.assertRaises(ValueError) as error:
            aframexr.Chart.from_dict(bad_data_specifications).to_html()
        self.assertEqual(str(error.exception), ERROR_MESSAGES['DATA_WITH_VALUES_AND_URL_IN_SPECS'])

    def test_data_has_not_field_data_url(self):
        """Verify that the error is raised when data has no field "data" or "url"."""
        bad_data_specifications = {'data': {}, 'mark': 'bar', 'encoding': {}}  # The same occurs with other marks
        with self.assertRaises(ValueError) as error:
            aframexr.Chart.from_dict(bad_data_specifications).to_html()
        self.assertEqual(str(error.exception), ERROR_MESSAGES['DATA_WITH_NOT_VALUES_NEITHER_URL_IN_SPECS'])

    def test_data_of_invalid_type(self):
        """Verify that the error is raised when data has invalid type."""
        err_data = 'invalid data type'
        with self.assertRaises(TypeError) as error:
            aframexr.Chart(err_data).to_html()
        self.assertEqual(
            str(error.exception),
            ERROR_MESSAGES['TYPE'].format(
                param_name='data', expected_type='Data or UrlData or DataFrame', current_type=type(err_data).__name__
            )
        )

    def test_mark_and_element_in_specs(self):
        """Verify that the error is raised when both "mark" and "element" are in chart specifications."""
        with self.assertRaises(ValueError) as error:
            aframexr.Chart.from_dict({'mark': '', 'element': ''}).to_html()
        self.assertEqual(str(error.exception), ERROR_MESSAGES['MARK_AND_ELEMENT_IN_SPECS'])

    def test_NULL_series(self):
        """Verify that the error is raised when having NULL series."""
        data = [{"id": 1, "value": None}, {"id": 2, "value": None}, {"id": 3, "value": None}]
        with self.assertRaises(ValueError) as error:
            aframexr.Chart(aframexr.Data(data)).mark_bar().encode(x='id', y='value').to_html()
        self.assertEqual(str(error.exception), 'Unknown dtype: Null.')

    def test_incorrect_equation_syntax(self):
        """Verify that the error is raised when equation has incorrect syntaxis."""
        invalid_equation = 'invalid_equation'
        with self.assertRaises(SyntaxError) as error:
            aframexr.Chart().transform_filter(invalid_equation)
        self.assertEqual(str(error.exception), 'Incorrect syntax, must be datum.{field} {operator} {value}')

    def test_add_error_not_isinstance_TopLevelMixin(self):
        """Verify that the error is raised when adding one chart to another thing."""
        one = aframexr.Chart(aframexr.UrlData(''))
        other = 2
        with self.assertRaises(TypeError) as error:
            one + other
        self.assertEqual(str(error.exception), f"Cannot add {type(other).__name__} to {type(one).__name__}.")

    def test_abstract_class_ElementCreator(self):
        """Verify that the error is raised when creating an ElementCreator object."""
        with self.assertRaises(TypeError) as error:
            aframexr.ElementCreator({})
        self.assertEqual(str(error.exception), 'ElementCreator is abstract')

    def test_moving_concat_chart_error(self):
        """Verify that the error is raised when moving one concatenated chart to another thing."""
        with self.assertRaises(ValueError) as error:
            (aframexr.Box() + aframexr.Box()).movable()
        self.assertEqual(str(error.exception), 'Concatenated charts cannot be movable.')

    def test_invalid_chart_specifications_not_data_having_mark(self):
        """Verify that the error is raised when chart specifications do not have field "data" having "mark"."""
        with self.assertRaises(ValueError) as error:
            aframexr.Chart.from_dict({'mark': ''}).show()  # Using show() to cover this method
        self.assertEqual(str(error.exception), ERROR_MESSAGES['DATA_NOT_IN_SPECS'])

    def test_invalid_chart_specifications_not_mark_element(self):
        """Verify that the error is raised when chart specifications do not have field "mark" or "element"."""
        with self.assertRaises(ValueError) as error:
            aframexr.Chart.from_dict({'neither_mark_or_element': ''}).to_html()
        self.assertEqual(str(error.exception), ERROR_MESSAGES['MARK_AND_ELEMENT_NOT_IN_SPECS'])

    def test_invalid_chart_type(self):
        """Verify that the error is raised when the chart type is invalid."""
        bad_chart_type = 'bad_mark'
        with self.assertRaises(ValueError) as error:
            aframexr.Chart.from_dict({'data': {'url': ''}, 'mark': bad_chart_type, 'encoding': ''}).to_html()
        self.assertEqual(str(error.exception), ERROR_MESSAGES['MARK_TYPE'].format(mark_type=bad_chart_type))

    def test_invalid_element_type(self):
        """Verify that the error is raised when the element type is invalid."""
        bad_element_type = 'bad_element_type'
        with self.assertRaises(ValueError) as error:
            aframexr.Chart.from_dict({'element': bad_element_type}).to_html()
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ELEMENT_TYPE'].format(element=bad_element_type))

    def test_save_invalid_type_format(self):
        """Verify that the error is raised when the save type is invalid."""
        bad_file_format = 'good_file.bad_format'
        with self.assertRaises(ValueError) as error:
            aframexr.Chart().save(bad_file_format)
        self.assertEqual(str(error.exception), 'Invalid file format. Must be "json" or "html"')



import aframexr
import unittest

from aframexr.utils.constants import AVAILABLE_ENVIRONMENTS

from tests.constants import *

URL = ('https://raw.githubusercontent.com/KhronosGroup/glTF-Sample-Assets/refs/heads/main/Models/AntiqueCamera/'
       'glTF/AntiqueCamera.gltf')


class TestMarkGLTFOK(unittest.TestCase):
    """Mark GLTF OK tests."""

    def test_simple(self):
        """Simple GLTF creation."""
        aframexr.GLTF(URL).to_html()

    def test_simple_with_pandas_not_installed(self):
        """Simple GLTF creation without having pandas installed."""
        import sys
        import importlib
        from unittest import mock

        sys.modules.pop('aframexr', None)
        sys.modules.pop('aframexr.api.components', None)

        with mock.patch.dict(sys.modules, {'pandas': None}):
            import aframexr
            importlib.reload(aframexr)

            aframexr.GLTF(URL).to_html()

            import aframexr.api.components as components
            self.assertIsNone(components.pd)
            self.assertIs(components.DataFrame, object)

    def test_from_json(self):
        """GLTF using from_json() method creation."""
        json_string = aframexr.GLTF(URL).to_json()
        gltf_chart = aframexr.Chart.from_json(json_string)
        gltf_chart.to_html()

        self.assertEqual(gltf_chart.to_json(), json_string)

    def test_position(self):
        """GLTF changing position creation."""
        for p in POSITIONS:
            aframexr.GLTF(URL, position=p).to_html()

    def test_position_format(self):
        """GLTF changing position format creation."""
        for p in POSITION_FORMATS:
            aframexr.GLTF(URL, position=p).to_html()

    def test_rotation(self):
        """GLTF changing rotation creation."""
        for r in ROTATIONS:
            aframexr.GLTF(URL, rotation=r).to_html()

    def test_rotation_format(self):
        """GLTF changing rotation format creation."""
        for r in ROTATION_FORMATS:
            aframexr.GLTF(URL, rotation=r).to_html()

    def test_scale(self):
        """GLTF changing scale creation."""
        for s in MARK_GLTF_SCALES:
            aframexr.GLTF(URL, scale=s).to_html()

    def test_position_rotation_scale(self):
        """GLTF changing position, rotation and scale creation."""
        for p, r, s in zip(POSITIONS, ROTATIONS, MARK_GLTF_SCALES):
            aframexr.GLTF(URL, position=p, rotation=r, scale=s).to_html()

    def test_concatenation(self):
        """GLTF concatenation creation."""
        concatenated_chart = aframexr.GLTF(URL, position=CONCATENATION_POSITIONS[0])
        for pos in CONCATENATION_POSITIONS[1:]:
            concatenated_chart += aframexr.GLTF(URL, position=pos)

        concatenated_chart.to_html()

    def test_environment(self):
        """Scene creation with personalized environment."""
        for e in AVAILABLE_ENVIRONMENTS:
            aframexr.GLTF(URL).to_html(environment=e)

    def test_save(self):
        """GLTF saving."""
        import tempfile
        from pathlib import Path

        with tempfile.TemporaryDirectory() as tmpdir:
            temp_html_file_path = Path(tmpdir) / "test.html"
            temp_json_file_path = Path(tmpdir) / "test.json"

            gltf = aframexr.GLTF(URL)
            gltf.save(str(temp_html_file_path))
            gltf.save(str(temp_json_file_path))

            assert temp_html_file_path.exists()
            assert temp_json_file_path.exists()

    def test_movable(self):
        """Movable GLTF creation."""
        html = aframexr.GLTF(URL).movable().to_html()
        self.assertIn('movable', html)



import aframexr
import unittest

from aframexr.utils.constants import AVAILABLE_ENVIRONMENTS, ERROR_MESSAGES
from .constants import *


class TestCylinderOK(unittest.TestCase):
    """Tests for simple cylinder creation."""

    def test_simple(self):
        """Simple cylinder creation."""
        cylinder = aframexr.Cylinder()
        cylinder.to_html()

    def test_from_dict(self):
        """Cylinder creation using form_dict() method."""
        cylinder = aframexr.Cylinder.from_dict(aframexr.Cylinder().to_dict())
        cylinder.to_html()

    def test_from_json(self):
        """Cylinder creation using form_json() method."""
        cylinder = aframexr.Cylinder.from_json(aframexr.Cylinder().to_json())
        cylinder.to_html()

    def test_color(self):
        """Cylinder with specific color creation."""
        for c in SIMPLE_ELEMENTS_COLORS:
            cylinder = aframexr.Cylinder(color=c)
            cylinder.to_html()

    def test_height(self):
        """Cylinder with specific height creation."""
        for h in ALL_MARK_DEPTHS_HEIGHTS_WIDTHS:
            cylinder = aframexr.Cylinder(height=h)
            cylinder.to_html()

    def test_position(self):
        """Cylinder with specific position creation."""
        for p in POSITIONS + POSITION_FORMATS:
            cylinder = aframexr.Cylinder(position=p)
            cylinder.to_html()

    def test_radius(self):
        """Cylinder with specific radius creation."""
        for r in RADIUS:
            cylinder = aframexr.Cylinder(radius=r)
            cylinder.to_html()

    def test_rotation(self):
        """Cylinder with specific rotation creation."""
        for r in ROTATIONS + ROTATION_FORMATS:
            cylinder = aframexr.Cylinder(rotation=r)
            cylinder.to_html()

    def test_all_parameters(self):
        """Cylinder creation with all parameters specified."""
        for c, h, p, ra, ro in zip(SIMPLE_ELEMENTS_COLORS, ALL_MARK_DEPTHS_HEIGHTS_WIDTHS, POSITIONS, RADIUS,
                                   ROTATIONS):
            cylinder = aframexr.Cylinder(color=c, height=h, position=p, radius=ra, rotation=ro)
            cylinder.to_html()

    def test_concatenation(self):
        """Concatenation of simple cylinders."""
        concatenated_chart = aframexr.Cylinder(position=CONCATENATION_POSITIONS[0])
        for p in CONCATENATION_POSITIONS[1:]:
            concatenated_chart += aframexr.Cylinder(position=p)
        concatenated_chart.to_html()

    def test_environment(self):
        """Scene creation with personalized environment."""
        for e in AVAILABLE_ENVIRONMENTS:
            aframexr.Cylinder().to_html(environment=e)

    def test_movable(self):
        """Movable cylinder creation."""
        html = aframexr.Cylinder().movable().to_html()
        self.assertIn('movable', html)


class TestCylinderERROR(unittest.TestCase):
    """Tests for cylinder creation error."""

    def test_color_error(self):
        """Cylinder with specific color creation error."""
        not_an_string = ['not a color']
        with self.assertRaises(TypeError) as error:
            # noinspection PyTypeChecker
            cylinder = aframexr.Cylinder(color=not_an_string)
            cylinder.to_html()
        self.assertEqual(
            str(error.exception),
            ERROR_MESSAGES['TYPE'].format(
                param_name='specs.color', expected_type=str.__name__, current_type=type(not_an_string).__name__
            )
        )

    def test_radius_error(self):
        """Cylinder with specific radius creation error."""
        for r in NOT_GREATER_THAN_0_RADIUS:
            with self.assertRaises(ValueError) as error:
                cylinder = aframexr.Cylinder(radius=r)
                cylinder.to_html()
            self.assertEqual(str(error.exception), ERROR_MESSAGES['POSITIVE_NUMBER'].format(param_name='specs.radius'))

    def test_environment_error(self):
        """Scene creation with invalid personalized environment."""
        environment = 'bad_environment'
        cylinder = aframexr.Cylinder()
        with self.assertRaises(ValueError) as error:
            # noinspection PyTypeChecker
            cylinder.to_html(environment=environment)
        self.assertEqual(str(error.exception), ERROR_MESSAGES['ENVIRONMENT'].format(environment=environment))



from abc import ABC, abstractmethod
import copy
import html
import json
import warnings

try:
    import pandas as pd
    from pandas import DataFrame
except ImportError:
    DataFrame = object
    pd = None

from IPython.display import display, HTML
from typing import Literal

from .aggregate import AggregatedFieldDef
from .data import Data, UrlData
from .encoding import Encoding, X, Y, Z
from .filters import FilterTransform
from .parameter import Parameter
from ..utils.scene_creator import SceneCreator
from ..utils.validators import AframeXRValidator


class TopLevelMixin:
    """Top level chart class."""

    def __init__(self, specs: dict):
        self._specifications = specs

    def _generate_iframe_html(self, ar_scale: str = None, environment: Literal['default', 'contact', 'egypt',
    'checkerboard', 'forest', 'goaland', 'yavapai', 'goldmine', 'arches', 'threetowers', 'poison', 'tron', 'japan',
    'dream', 'volcano', 'starry', 'osiris'] = 'default'):
        return (
            '<iframe '
            f'srcdoc="{html.escape(self.to_html(ar_scale=ar_scale, environment=environment), quote=True)}" '  # Raw HTML
            'width="100%" '  # Adjust to maximum width
            'height="400" '  # Height of the iframe
            'style="border:none;" '
            'sandbox="allow-scripts allow-forms allow-same-origin" '
            'loading="lazy" '  # For optimization
            '></iframe>'
        )

    def _resolve_data(self):
        """Resolves the data reference of the specifications."""
        def materialize_data(data):
            AframeXRValidator.validate_type(
                'data', data, (Data, UrlData, DataFrame)  # type: ignore[arg-type] --> DataFrame
            )

            if isinstance(data, Data):
                return {'values': data.values}
            elif isinstance(data, UrlData):
                return {'url': data.url}
            elif pd is not None and isinstance(data, pd.DataFrame):
                arr = data.to_numpy()  # type: ignore --> because of DataFrame
                cols = data.columns.tolist()  # type: ignore --> because of DataFrame
                return {'values': [dict(zip(cols, row)) for row in arr]}
            else:  # pragma: no cover (AframeXRValidator.validate_type() should have validate data type)
                raise RuntimeError('Unreachable code: AframeXRValidator.validate_type() should have validate data type')

        for specs in self._specifications.get('concat', [self._specifications]):
            if 'data_ref' in specs:
                specs['data'] = materialize_data(specs.pop('data_ref'))

    def _repr_html_(self):  # pragma: no cover (as this method is called in notebooks)
        """Returns the iframe HTML for showing the scene in the notebook."""
        return self._generate_iframe_html()

    # Concatenating charts
    def __add__(self, other):
        """
        Concatenation of charts (place charts in the same scene).
        Creates and returns a new scene with the charts. The original charts are not modified.
        """
        if not isinstance(other, TopLevelMixin):
            raise TypeError(f"Cannot add {type(other).__name__} to {type(self).__name__}.")

        self_specs_list = copy.deepcopy(self._specifications.get('concat', [self._specifications]))
        other_specs_list = copy.deepcopy(other._specifications.get('concat', [other._specifications]))

        new = self.copy()  # Create a copy to modify
        new._specifications = {'concat': self_specs_list + other_specs_list}
        return new

    # Copy of the chart
    def __deepcopy__(self, memo):
        """Optimized deepcopy avoiding copying large data values."""
        if id(self) in memo:  # Avoid copying several times the same object
            return memo[id(self)]

        new_instance = self.__class__.__new__(self.__class__)
        memo[id(self)] = new_instance

        # Separate data reference from the rest of the specifications
        specs_copy = self._specifications.copy()  # Shallow copy
        data_ref = specs_copy.pop('data_ref', None)

        # Deep copy of the rest
        new_instance._specifications = copy.deepcopy(specs_copy, memo)

        # Restore data reference
        if data_ref is not None:
            new_instance._specifications['data_ref'] = data_ref

        return new_instance

    def copy(self):
        """Return a deep copy of the chart while keeping large data as reference."""
        return copy.deepcopy(self)

    # Importing charts
    @staticmethod
    def from_dict(specs: dict) -> 'TopLevelMixin':
        """
        Import the chart from the JSON dict specifications.

        Parameters
        ----------
        specs : dict
            JSON specifications of the chart.

        Raises
        ------
        TypeError
            If specs is not a dictionary.
        """
        AframeXRValidator.validate_type('specs', specs, dict)
        chart = Chart()
        chart._specifications = copy.deepcopy(specs)
        return chart

    @staticmethod
    def from_json(specs: str) -> 'TopLevelMixin':
        """
        Create the chart from the JSON string specifications.

        Parameters
        ----------
        specs : str
            JSON specifications of the chart.

        Raises
        ------
        TypeError
            If specs is not a string.
        JSONDecodeError
            If specs is not a valid JSON.
        """
        AframeXRValidator.validate_type('specs', specs, str)
        chart = Chart()
        chart._specifications = copy.deepcopy(json.loads(specs))
        AframeXRValidator.validate_chart_specs(chart._specifications)
        return chart

    # Movable
    def movable(self):
        """
        Make the entity movable.

        Concatenated charts cannot be movable, each chart must be defined as movable before concatenating.
        """
        self_copy = self.copy()
        if 'concat' in self_copy._specifications:
            raise ValueError('Concatenated charts cannot be movable.')

        self_copy._specifications['movable'] = True
        return self_copy

    # Exporting charts
    def save(self, fp: str, file_format: Literal['json', 'html'] = None, environment: Literal['default', 'contact',
    'egypt', 'checkerboard', 'forest', 'goaland', 'yavapai', 'goldmine', 'arches', 'threetowers', 'poison', 'tron',
    'japan', 'dream', 'volcano', 'starry', 'osiris'] = 'default'):
        """
        Saves the chart into a file, supported formats are JSON and HTML.

        Parameters
        ----------
        fp : str
            File path.
        file_format : str (optional)
            Format of the file could be ['html', 'json'].
            If no format is specified, the chart will be saved depending on the file extension.
        environment : str (optional)
            Environment of the scene.

        Raises
        ------
        ValueError
            If file_format is invalid.
        """
        AframeXRValidator.validate_type('fp', fp, str)
        self_copy = self.copy()

        if file_format == 'html' or fp.endswith('.html'):
            with open(fp, 'w') as file:
                file.write(self_copy.to_html(environment=environment))
        elif file_format == 'json' or fp.endswith('.json'):
            with open(fp, 'w') as file:
                specs = self_copy.to_dict()
                specs['environment'] = environment
                AframeXRValidator.validate_chart_specs(specs)
                json.dump(specs, file, indent=4)
        else:
            raise ValueError('Invalid file format. Must be "json" or "html"')

    # Showing the scene
    def show(self, ar_scale: str = None, environment: Literal['default', 'contact', 'egypt', 'checkerboard', 'forest',
    'goaland', 'yavapai', 'goldmine', 'arches', 'threetowers', 'poison', 'tron', 'japan', 'dream', 'volcano', 'starry',
    'osiris'] = 'default'):
        """Show the scene in the notebook."""
        with warnings.catch_warnings():
            # Do not show the warning --> UserWarning: Consider using IPython.display.IFrame instead
            warnings.filterwarnings('ignore', message='Consider using IPython.display.IFrame instead')

            self_copy = self.copy()
            self_copy._resolve_data()

            html_obj = HTML(self_copy._generate_iframe_html(
                ar_scale=ar_scale,
                environment=environment
            ))

            display(html_obj)
            return html_obj

    # Chart formats
    def to_dict(self) -> dict:
        """Returns the scene specifications as a dictionary."""
        self_copy = self.copy()
        self_copy._resolve_data()

        AframeXRValidator.validate_chart_specs(self_copy._specifications)
        return self_copy._specifications

    def to_html(self, ar_scale: str = None, environment: Literal['default', 'contact', 'egypt', 'checkerboard',
    'forest', 'goaland', 'yavapai', 'goldmine', 'arches', 'threetowers', 'poison', 'tron', 'japan', 'dream', 'volcano',
    'starry', 'osiris'] = 'default') -> str:
        """Returns the HTML representation of the scene."""
        self_copy = self.copy()
        self_copy._resolve_data()

        if ar_scale is not None: self_copy._specifications['ar_scale'] = ar_scale
        self_copy._specifications['environment'] = environment
        AframeXRValidator.validate_chart_specs(self_copy._specifications)
        return SceneCreator.create_scene(self_copy._specifications)

    def to_json(self) -> str:
        """Returns the JSON string of the scene."""
        return json.dumps(self.to_dict())  # Method to_dict() validates chart specifications


class Chart(TopLevelMixin):
    """
    Simple chart class.

    Parameters
    ----------
    data : Data | UrlData | DataFrame
        Data, UrlData object or pandas DataFrame of the data.
    depth : float (optional)
        Depth of the chart. If not defined, using DEFAULT_CHART_DEPTH.
    height : float (optional)
        Height of the chart. If not defined, using DEFAULT_CHART_HEIGHT.
    position : str (optional)
        Position of the chart. The format is: 'x y z'. Refers to the position for the origin of coordinate system.
        If not defined, using DEFAULT_CHART_POS.
    rotation : str (optional)
        Rotation of the chart in degrees. The format is: 'x y z'. The rotation axis is the coordinate system.
        If not defined, using DEFAULT_CHART_ROTATION.
    width : float (optional)
        Width of the chart. If not defined, using DEFAULT_CHART_WIDTH.

    Raises
    ------
    TypeError
        If any parameter has invalid type.
    ValueError
        If depth, height, position, rotation or width is invalid.
    """
    def __init__(self, data: Data | UrlData | DataFrame = None, depth: float = None, height: float = None,
                 title: str = None, position: str = None, rotation: str = None, width: float = None):
        super().__init__({})  # Initiate specifications

        if data is not None: self._specifications['data_ref'] = data
        if position is not None: self._specifications['position'] = position
        if rotation is not None: self._specifications['rotation'] = rotation
        if depth is not None: self._specifications['depth'] = depth
        if height is not None: self._specifications['height'] = height
        if width is not None: self._specifications['width'] = width
        if title is not None: self._specifications['title'] = title

    # Parameters
    def add_params(self, *params: Parameter):
        self_copy = self.copy()

        for p in params:
            AframeXRValidator.validate_type('params', p, Parameter)

        if 'params' not in self_copy._specifications:
            self_copy._specifications['params'] = []

        self_copy._specifications['params'].extend(
            [p.to_specs() for p in params]
        )

        return self_copy

    # Types of charts
    def mark_arc(self, radius: float = None):
        """
        Pie chart and doughnut chart.

        Parameters
        ----------
        radius : float (optional)
            Outer radius of the pie chart. If not specified, using DEFAULT_PIE_RADIUS. Must be greater than 0.
        """
        self_copy = self.copy()

        self_copy._specifications['mark'] = {'type': 'arc'}

        if radius is not None:
            AframeXRValidator.validate_positive_number('radius', radius)
            self_copy._specifications['mark']['radius'] = radius

        return self_copy

    def mark_bar(self, color: str = None, size: float = None):
        """
        Bars chart.

        Parameters
        ----------
        color: str (optional)
            Color of the bars. If not defined, using DEFAULT_ELEMENTS_COLOR_IN_CHART.
        size : float (optional)
            Width of the bars. If not specified, bars will be adjusted automatically. Must be greater than 0.

        Raises
        ------
        ValueError
            If defined size is not greater than 0.
        """
        self_copy = self.copy()

        self_copy._specifications['mark'] = {'type': 'bar'}

        if color is not None:
            AframeXRValidator.validate_type('color', color, str)
            self_copy._specifications['mark']['color'] = color

        if size is not None:
            AframeXRValidator.validate_positive_number('size', size)
            self_copy._specifications['mark']['size'] = size

        return self_copy

    def mark_line(self, color: str = None, point: bool = None):
        """
        Line chart.

        Parameters
        ----------
        color: str (optional)
            Color of the line. If not defined, using DEFAULT_ELEMENTS_COLOR_IN_CHART.
        point : bool (optional)
            Either if add points in vertices or not. If not defined, using
        """
        self_copy = self.copy()

        self_copy._specifications['mark'] = {'type': 'line'}

        if color is not None:
            AframeXRValidator.validate_type('color', color, str)
            self_copy._specifications['mark']['color'] = color

        if point is not None:
            AframeXRValidator.validate_type('point', point, bool)
            self_copy._specifications['mark']['point'] = point

        return self_copy

    def mark_point(self, color: str = None, size: float = None):
        """
        Scatter plot and bubble chart.

        Parameters
        ----------
        color: str (optional)
            Color of the spheres. If not defined, using DEFAULT_ELEMENTS_COLOR_IN_CHART.
        size : float (optional)
            Maximum radius of the point. If not specified, using DEFAULT_POINT_RADIUS. Must be greater than 0.

        Raises
        ------
        ValueError
            If size is not greater than 0.
        """
        self_copy = self.copy()

        self_copy._specifications['mark'] = {'type': 'point'}

        if color is not None:
            AframeXRValidator.validate_type('color', color, str)
            self_copy._specifications['mark']['color'] = color

        if size is not None:
            AframeXRValidator.validate_positive_number('size', size)
            self_copy._specifications['mark']['size'] = size

        return self_copy

    # Parameters of the chart
    def encode(self, color: str = None, size: str = None, theta: str = None, x: str | X = None, y: str | Y = None,
               z: str | Z = None):
        """
        Add properties to the chart.

        Parameters
        ----------
        color : str (optional)
            Field of the data that will determine the color of the elements in the chart.
        size : str (optional)
            Field of the data that will determine the size of the sphere in the bubble chart (must be quantitative).
        theta : str (optional)
            Field of the data that will determine the arcs of the pie chart (must be quantitative).
        x : str | X (optional)
            Field of the data that will determine the x-axis of the chart.
        y : str | Y (optional)
            Field of the data what will determine the y-axis of the chart.
        z : str | Z (optional)
            Field of the data what will determine the z-axis of the chart.

        Raises
        ------
        TypeError
            If the encoding type is incorrect.
        ValueError
            If the encoding values are incorrect.
        """
        filled_params = {}  # Dictionary that will store the parameters that have been filled

        # Verify the type of the arguments and store the filled parameters
        if color is not None:
            AframeXRValidator.validate_type('color', color, str)
            filled_params['color'] = color
        if size is not None:
            AframeXRValidator.validate_type('size', size, str)
            filled_params['size'] = size
        if theta is not None:
            AframeXRValidator.validate_type('theta', theta, str)
            filled_params['theta'] = theta
        if x is not None:
            AframeXRValidator.validate_type('x', x, (str, X))
            filled_params['x'] = x
        if y is not None:
            AframeXRValidator.validate_type('y', y, (str, Y))
            filled_params['y'] = y
        if z is not None:
            AframeXRValidator.validate_type('z', z, (str, Z))
            filled_params['z'] = z

        # Do the encoding
        self_copy = self.copy()

        encoding = self_copy._specifications.setdefault('encoding', {})  # For merging possible further encodings
        for param_key in filled_params:
            param_value = filled_params[param_key]
            if isinstance(param_value, Encoding):
                encoding.update(param_value.to_dict())
            else:
                formula, encoding_type = Encoding.split_field_and_encoding(param_value)
                field, aggregate_op = AggregatedFieldDef.split_operator_field(formula)
                encoding[param_key] = {'field': field}
                if aggregate_op:
                    encoding[param_key]['aggregate'] = aggregate_op
                if encoding_type:
                    encoding[param_key]['type'] = encoding_type

        return self_copy

    def properties(self, data: Data | UrlData | DataFrame = None, depth: str = None, height: str = None,
                   position: str = None, rotation: str = None, title: str = None, width: str = None):
        """Modify general properties of the chart."""
        self_copy = self.copy()

        if data is not None: self_copy._specifications['data_ref'] = data
        if position is not None: self_copy._specifications['position'] = position
        if rotation is not None: self_copy._specifications['rotation'] = rotation
        if depth is not None: self_copy._specifications['depth'] = depth
        if height is not None: self_copy._specifications['height'] = height
        if width is not None: self_copy._specifications['width'] = width
        if title is not None: self_copy._specifications['title'] = title

        return self_copy

    # Modifying data
    def transform_aggregate(self, groupby: list = None, **kwargs):
        """
        Aggregates the data with the specified aggregate function, grouped by the specified groupby.

        Parameters
        ----------
        groupby : list | None
            Data fields that will be grouped, optional. If not set, the defined fields in encode() method will be taken.
        kwargs : dict
            Format is: <new_field>=<aggregate_op>(<data_field>).
        """
        AframeXRValidator.validate_type('groupby', groupby, (list, type(None)))

        # Create a copy of the chart (in case of assignation, to preserve the main chart)
        aggreg_chart = self.copy()

        aggregates_to_dict = []
        for as_field, aggregate_formula in kwargs.items():
            field, aggregate_op = AggregatedFieldDef.split_operator_field(str(aggregate_formula))
            aggregate_object = AggregatedFieldDef(aggregate_op, field, as_field)
            aggregates_to_dict.append(aggregate_object.to_dict())

        aggregate_specs = {'aggregate': aggregates_to_dict}
        if groupby:
            aggregate_specs['groupby'] = groupby

        if not aggreg_chart._specifications.get('transform'):  # First time filtering the chart (create field)
            aggreg_chart._specifications['transform'] = [aggregate_specs]
        else:  # Not the first filter of the chart (add to aggregates)
            aggreg_chart._specifications['transform'].append(aggregate_specs)
        return aggreg_chart

    def transform_filter(self, equation_filter: str | FilterTransform | Parameter):
        """
        Filters the chart with the given transformation.

        Parameters
        ----------
        equation_filter : str | FilterTransform | Parameter
            The equation string of the filter transformation, a Filter object (see Examples) or Parameter object.

        Raises
        ------
        TypeError
            If equation is not a string, a Filter object or a Parameter object.

        Notes
        -----
        Can be concatenated with the rest of functions of the Chart, without needing an asignation. It can also be
        concatenated several times (the result will be an addition of the filters, in order of assignation).

        Examples
        --------
        *Using transform_filter() giving the equation string:*

        >>> import aframexr
        >>> data = aframexr.UrlData('./data.json')
        >>> chart = aframexr.Chart(data).mark_bar().encode(x='model', y='sales')
        >>> filtered_chart = chart.transform_filter('datum.motor == diesel')
        >>> #filtered_chart.show()

        *Using transform_filter() giving a Filter object*

        >>> import aframexr
        >>> data = aframexr.UrlData('./data.json')
        >>> chart = aframexr.Chart(data).mark_bar().encode(x='model', y='sales')
        >>> filter_object = aframexr.FieldEqualPredicate(field='motor', equal='diesel')
        >>> filtered_chart = chart.transform_filter(filter_object)
        >>> #filtered_chart.show()
        """
        # Validate the type of equation_filter and get a filter object from the equation_filter
        AframeXRValidator.validate_type(
            'equation_filter', equation_filter, (str, FilterTransform, Parameter)
        )
        if isinstance(equation_filter, str):
            filter_transform = FilterTransform.from_equation(equation_filter)
        elif isinstance(equation_filter, (FilterTransform, Parameter)):
            filter_transform = equation_filter
        else:  # pragma: no cover
            raise RuntimeError('Unreachable code. Parameter should have been validated before')

        # Create a copy of the chart (in case of assignation, to preserve the main chart)
        filt_chart = self.copy()

        # Add the information of the filter object to the specifications
        if not filt_chart._specifications.get('transform'):  # First time filtering the chart
            filt_chart._specifications['transform'] = [{'filter': filter_transform.to_dict()}]
        else:  # Not the first filter of the chart
            filt_chart._specifications['transform'].append({'filter': filter_transform.to_dict()})
        return filt_chart  # Returns the copy of the chart


class Element(TopLevelMixin, ABC):
    @abstractmethod
    def __init__(self, **kwargs):
        element_name = self.__class__.__name__.lower()  # Use child class's name for defining the element
        super().__init__({'element': element_name})

        self._specifications.update(
            {key: value for key, value in kwargs.items()
             if value is not None}
        )


# Single elements
class Box(Element):
    def __init__(self, color: str = None, depth: float = None, height: float = None, position: str = None,
                 rotation: str = None, width: float = None):
        super().__init__(color=color, depth=depth, height=height, position=position, rotation=rotation, width=width)


class Cone(Element):
    def __init__(self, color: str = None, height: float = None, position: str = None, radius_bottom: float = None,
                 radius_top: float = None):
        super().__init__(color=color, height=height, position=position, radius_bottom=radius_bottom,
                         radius_top=radius_top)


class Cylinder(Element):
    def __init__(self, color: str = None, height: float = None, position: str = None, radius: float = None,
                 rotation: str = None):
        super().__init__(color=color, height=height, position=position, radius=radius, rotation=rotation)


class Dodecahedron(Element):
    def __init__(self, color: str = None, position: str = None, radius: float = None):
        super().__init__(color=color, position=position, radius=radius)


class GLTF(Element):
    def __init__(self, src: str, scale: str = None, position: str = None, rotation: str = None):
        super().__init__(src=src, scale=scale, position=position, rotation=rotation)


class Icosahedron(Element):
    def __init__(self, color: str = None, position: str = None, radius: float = None):
        super().__init__(color=color, position=position, radius=radius)


class Image(Element):
    def __init__(self, src: str, height: float = None, position: str = None, rotation: str = None, width: float = None):
        super().__init__(src=src, height=height, position=position, rotation=rotation, width=width)


class Line(Element):
    def __init__(self, start: str, end: str, color: str = None):
        super().__init__(start=start, end=end, color=color)


class Octahedron(Element):
    def __init__(self, color: str = None, position: str = None, radius: float = None):
        super().__init__(color=color, position=position, radius=radius)


class Plane(Element):
    def __init__(self, color: str = None, height: float = None, position: str = None, rotation: str = None,
                 width: float = None):
        super().__init__(color=color, height=height, position=position, rotation=rotation, width=width)


class Sphere(Element):
    def __init__(self, color: str = None, position: str = None, radius: float = None):
        super().__init__(color=color, position=position, radius=radius)


class Tetrahedron(Element):
    def __init__(self, color: str = None, position: str = None, radius: float = None):
        super().__init__(color=color, position=position, radius=radius)


class Text(Element):
    def __init__(self, value: str, align: Literal['center', 'left', 'right'] = None, color: str = None,
                 position: str = None, rotation: str = None, scale: str = None):
        super().__init__(align=align, color=color, position=position, rotation=rotation, value=value, scale=scale)


class Torus(Element):
    def __init__(self, color: str = None, position: str = None, radius: float = None, radius_tubular: float = None,
                 rotation: str = None):
        super().__init__(color=color, position=position, radius=radius, radius_tubular=radius_tubular,
                         rotation=rotation)



from .entities_html_creator import ChartsHTMLCreator

HTML_SCENE_TEMPLATE = """<!DOCTYPE html>
<html lang="en">
<head>
  <title>VR Scene</title>
  <meta charset="utf-8">
  <script src="https://aframe.io/releases/1.7.1/aframe.min.js"></script>
  <script src="https://cdn.jsdelivr.net/gh/c-frame/aframe-extras@7.6.1/dist/aframe-extras.min.js"></script>
  <script src="https://unpkg.com/aframe-environment-component@1.5.0/dist/aframe-environment-component.min.js"></script>
  <script src="https://cdn.jsdelivr.net/gh/davidlab20/TFG@v0.10.2/docs/static/scripts/main.min.js"></script>
</head>
<body>
<a-scene cursor="rayOrigin: mouse" raycaster="objects: [raycastable]" drag-controls="mode: cursor"
         xr-mode-ui="XRMode: xr" webxr="optionalFeatures: hit-test, local-floor;" ar-hit-test="target: #charts">
  <a-entity id="user" movement-controls="speed: 0.1">

    <!-- Camera -->
    <a-camera id="camera" position="0 1.6 0" active="true" wasd-controls="acceleration: 15"></a-camera>

    <!-- VR controllers -->
    <a-entity laser-controls="hand: right" raycaster="objects: [raycastable]" drag-controls="mode: vr"></a-entity>
    <a-entity laser-controls="hand: left" raycaster="objects: [raycastable]" drag-controls="mode: vr"></a-entity>
  </a-entity>

  <!-- HUD -->
  <a-entity id="HUD" visible="false" scale-on-enter-ar>
    <a-plane id="HUD-plane" width="2" shader="flat" color="grey"></a-plane>
    <a-entity id="HUD-texts"></a-entity>
  </a-entity>

  <!-- Environment -->
  <a-entity environment="preset: {environment}" hide-on-enter-ar></a-entity>
  <a-light type="ambient" show-on-enter-ar></a-light>

  <!-- Elements -->
  <a-entity id="charts" scale-on-enter-ar{ar_scale_value} look-at-camera-on-ar>
    {elements}
  </a-entity>
</a-scene>
</body>
</html>"""


class SceneCreator:
    @staticmethod
    def create_scene(specs: dict):
        """
        Creates the HTML scene from the JSON specifications.

        Parameters
        ----------
        specs : dict
            Specifications of the elements composing the scene.
        """
        ar_scale = specs.get('ar_scale')
        if ar_scale is None:
            ar_scale_value = ''  # Using default's JavaScript value
        else:
            ar_scale_value = f'="scale: {ar_scale}"'  # Setting value for JavaScript
        environment = specs.get('environment', 'default')
        elements_html = ChartsHTMLCreator.create_charts_html(specs)
        return HTML_SCENE_TEMPLATE.format(
            ar_scale_value=ar_scale_value, environment=environment, elements=elements_html
        )
